% *** ACP Algorithm ***
% *VB algorithm with Adaptive Correlated Prior
%  to solve set of equations y = Mx + n, where x is hight-resolution image
% *see paper: J. evk, V. mdl and F. roubek, An Adaptive Correlated
% Image Prior for Image Restoration Problems, IEEE signal processing
% letters, 2018.
%  
% INPUT:
%  K ....... number of low-resolution images
%  yk ...... cell array of K low-resolution (LR) images (cell 1 x K)
%             each cell contains one LR images (matrix of size p1 x p2)
%  Mk ...... cell arary of K degradation matrix  (cell 1 x K)
%             each cell contains one degradation matrix (of size p1*p2 x N)
%  n1,n2 ... size of high-resolution image (n1 x n2: important for visulation)
%
% written by Jakub Sevcik, Vaclav Smidl and Filip Sroubek 2018
% for more information contact: jsevcik@rice.zcu.cz

%% data preparing
load data_example

M = [];
y = [];
for i = 1:K
    M = [M;Mk{i}];
    y = [y;yk{i}(:)];
end

[p, N] = size(M);
[p1, p2] = size(yk{1});

%% preparing of auxiliary structures
indL = zeros(N,N);
indDiag = [1 1 zeros(1,n1-2) 1 zeros(1,n1*(n2-1)-1)];
for j = 1:(N)
    if indDiag(j)
        indL = indL + diag(ones(N-j+1,1),j-1);
    end
end
for j = 1:(n2-1)
    indL(n1*j,n1*j+1) = 0;
end   
indLD = (logical(indL'));
indL = (logical(indL' - eye(N)));
sum_indL = sum(sum(indL));
clear indDiag

cell_sigma_lj = cell(N-1,1);
for j = 1:N-1
    cell_sigma_lj{j} = zeros(sum(indL(:,j)));
end

aux_LULt = sparse(zeros*eye(N));
aux_LtxxtL = ones(N,1);

MtM = M'*M;
for i = 1:1:K
    MktMk{i} = (Mk{i})'*Mk{i};
end
Mty = M'*y;
ytM = y'*M;
yty = y'*y;

%% inicialization of parameters and hyper-parameters
hat_beta = ones(K,1);
hat_alpha = 1e-5*ones(N,1);

hat_psi = (1e-5*indL);

koef_l0 = -0.5;
L_0 = sparse(zeros(N,N) + indL*koef_l0);
hat_L = (L_0+eye(N));

gamma_0_alpha = 1e-8;
delta_0_alpha = 1e-8;

gamma_0_beta = 1e-8;
delta_0_beta = 1e-8;

zeta_0 = 1e-2;
eta_0 = 1e-2;

iterations = 100;

%% computation
for iter = 1:1:iterations    
    % ------ x ------
    sigma_x = inv(  M'*sparse(diag(kron(hat_beta,ones(p1*p2,1))))*M   +  ...
                    hat_L*sparse(diag(hat_alpha))*hat_L' + aux_LULt  );
    
    hat_x = sigma_x*( M'*(kron(hat_beta,ones(p1*p2,1)).*y) );    
    
    hat_x(hat_x<0) = 0;     % positivity of x
    hat_xxt = hat_x*hat_x' + sigma_x;   

    % ------ alpha ------
    gamma_alpha = gamma_0_alpha + 0.5;
    for j = 1:(N-1)
       aux_LtxxtL(j) = hat_L(indLD(:,j),j)'*hat_xxt(indLD(:,j),indLD(:,j))*hat_L(indLD(:,j),j) + ...
                       trace(  cell_sigma_lj{j}*( hat_xxt(indL(:,j),indL(:,j)) )  );
    end
    aux_LtxxtL(N) = hat_xxt(N,N);
    delta_alpha = delta_0_alpha + 0.5*aux_LtxxtL;
    
    hat_alpha = gamma_alpha./delta_alpha;
    
    % ------ l, psi ------
    aux_LULt = (zeros(N)); 
    zeta = zeta_0 + 0.5;
    for j = 1:(N-1)
        sigma_lj = (  hat_alpha(j)*hat_xxt(indL(:,j),indL(:,j)) + diag(hat_psi(indL(:,j),j))  )^(-1);
        hat_lj = sigma_lj*( -hat_alpha(j)*hat_xxt(indL(:,j),j) + hat_psi(indL(:,j),j).*L_0(indL(:,j),j));
 
        hat_L(indL(:,j),j) = hat_lj;
        cell_sigma_lj{j} = sigma_lj;
        
        eta = eta_0 + 0.5*(hat_lj.^2 + diag(sigma_lj)) - L_0(indL(:,j),j).*hat_lj + 0.5*L_0(indL(:,j),j).*L_0(indL(:,j),j);   
        hat_psi(indL(:,j),j) = zeta./eta;

        aux_LULt(indL(:,j),indL(:,j)) = aux_LULt(indL(:,j),indL(:,j)) + sigma_lj*hat_alpha(j);
    end
      
    % ------ beta ------
    gamma_beta = gamma_0_beta + 0.5*(p1*p2);
    for i=1:K
        delta_beta = delta_0_beta + 0.5*sum(sum(hat_xxt.*MktMk{i})) - (yk{i}(:))'*Mk{i}*hat_x + 0.5*(yk{i}(:))'*yk{i}(:);
        hat_beta(i) = gamma_beta/delta_beta;
    end
    
    if mod(iter,10) == 0
        disp(['ACP alg.: iteration ' num2str(iter) '/' num2str(iterations)])    
    end
end

%% visualisation of the reconstruction
figure
imagesc(reshape(hat_x,n1,n2))
title('reconstruction of high-resolution image')
colormap gray
axis off equal